#include "ivm.h"
#include <map>
#include <mutex>
#include <cstring>
#include "urpc.h"
static std::map<device_t, urpc_device_handle_t> impl_by_handle;
static std::mutex impl_by_handle_mutex;
static void push_data(uint8_t **where, const void *data, size_t size)
{
    memcpy(*where, data, size);
    *where += size;
}
static void push_float(uint8_t **where, float value)
{
    push_data(where, &value, sizeof(value));
}
static void push_double(uint8_t **where, double value)
{
    push_data(where, &value, sizeof(value));
}
static float pop_float(uint8_t **where)
{
    float result;
    memcpy(&result, *where, sizeof(result));
    *where += sizeof(result);
    return result;
}
static double pop_double(uint8_t **where)
{
    double result;
    memcpy(&result, *where, sizeof(result));
    *where += sizeof(result);
    return result;
}
#define GENERATE_PUSH(Type) \
static void push_##Type(uint8_t **where, Type value) { \
    push_data(where, &value, sizeof(value)); \
}
GENERATE_PUSH(uint64_t)
GENERATE_PUSH(uint32_t)
GENERATE_PUSH(uint16_t)
GENERATE_PUSH(uint8_t)
GENERATE_PUSH(int64_t)
GENERATE_PUSH(int32_t)
GENERATE_PUSH(int16_t)
GENERATE_PUSH(int8_t)
#define GENERATE_POP(Type) \
static Type pop_##Type(uint8_t **where) { \
    Type result; \
    memcpy(&result, *where, sizeof(result)); \
    *where += sizeof(result); \
    return (Type)result; \
}
GENERATE_POP(uint64_t)
GENERATE_POP(uint32_t)
GENERATE_POP(uint16_t)
GENERATE_POP(uint8_t)
GENERATE_POP(int64_t)
GENERATE_POP(int32_t)
GENERATE_POP(int16_t)
GENERATE_POP(int8_t)
device_t ivm_open_device(const char* uri)
{
	device_t handle;
	urpc_device_handle_t device;
	device = urpc_device_create(uri);
	if(device == NULL) {
	    return device_undefined;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    do
	    {
	        handle = rand();
	    }
	    while(impl_by_handle.count(handle) != 0);
	    impl_by_handle[handle] = device;
	}
	return handle;
}

device_t ivm_libversion(char* lib_version)
{
	char *lib_v = "1.0.2";
	strcpy(lib_version,lib_v);
	return result_ok;
}

result_t ivm_get_measurement(device_t handle, ivm_in_get_measurement_t *input, ivm_out_get_measurement_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t in_buffer[2];
	memset(in_buffer, 0, 2);
	uint8_t out_buffer[200];
	memset(out_buffer, 0, 200);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	p = in_buffer;
	push_uint16_t(&p, input->FrameNumber);
	if(urpc_device_send_request(device, "msre", in_buffer, 2, out_buffer, 200) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	for(i=0; i<25; i++) output->Voltage[i] = pop_float(&p);
	for(i=0; i<25; i++) output->Current[i] = pop_float(&p);
	return result_ok;
}

result_t ivm_get_identity_information(device_t handle, ivm_get_identity_information_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t out_buffer[72];
	memset(out_buffer, 0, 72);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "ginf", NULL, 0, out_buffer, 72) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	for(i=0; i<16; i++) output->Manufacturer[i] = pop_uint8_t(&p);
	for(i=0; i<16; i++) output->ProductName[i] = pop_uint8_t(&p);
	for(i=0; i<16; i++) output->ControllerName[i] = pop_uint8_t(&p);
	output->HardwareMajor = pop_uint8_t(&p);
	output->HardwareMinor = pop_uint8_t(&p);
	output->HardwareBugfix = pop_uint16_t(&p);
	output->BootloaderMajor = pop_uint8_t(&p);
	output->BootloaderMinor = pop_uint8_t(&p);
	output->BootloaderBugfix = pop_uint16_t(&p);
	output->FirmwareMajor = pop_uint8_t(&p);
	output->FirmwareMinor = pop_uint8_t(&p);
	output->FirmwareBugfix = pop_uint16_t(&p);
	output->SerialNumber = pop_uint32_t(&p);
	for(i=0; i<8; i++) output->Reserved[i] = pop_uint8_t(&p);
	return result_ok;
}

result_t ivm_start_autocalibration(device_t handle, ivm_start_autocalibration_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t out_buffer[1];
	memset(out_buffer, 0, 1);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "clbr", NULL, 0, out_buffer, 1) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	output->Result = pop_uint8_t(&p);
	return result_ok;
}

result_t ivm_get_status(device_t handle, ivm_get_status_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t out_buffer[54];
	memset(out_buffer, 0, 54);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "gets", NULL, 0, out_buffer, 54) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	output->Status = pop_uint32_t(&p);
	output->Temp = pop_int16_t(&p);
	for(i=0; i<48; i++) output->Reserved[i] = pop_uint8_t(&p);
	return result_ok;
}

result_t ivm_start_measurement(device_t handle)
{
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "smre", NULL, 0, NULL, 0) != urpc_result_ok)
	{
	    return result_error;
	}
	return result_ok;
}

result_t ivm_check_measurement_status(device_t handle, ivm_check_measurement_status_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t out_buffer[16];
	memset(out_buffer, 0, 16);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "mrdy", NULL, 0, out_buffer, 16) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	output->ReadyStatus = pop_uint8_t(&p);
	for(i=0; i<15; i++) output->Reserved[i] = pop_uint8_t(&p);
	return result_ok;
}

result_t ivm_get_measurement_raw(device_t handle, ivm_in_get_measurement_raw_t *input, ivm_out_get_measurement_raw_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t in_buffer[2];
	memset(in_buffer, 0, 2);
	uint8_t out_buffer[100];
	memset(out_buffer, 0, 100);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	p = in_buffer;
	push_uint16_t(&p, input->Frame);
	if(urpc_device_send_request(device, "msrr", in_buffer, 2, out_buffer, 100) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	for(i=0; i<25; i++) output->ADCHighCode[i] = pop_uint16_t(&p);
	for(i=0; i<25; i++) output->ADCLowCode[i] = pop_uint16_t(&p);
	return result_ok;
}

result_t ivm_get_device_rank(device_t handle, ivm_get_device_rank_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t out_buffer[132];
	memset(out_buffer, 0, 132);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "gdrk", NULL, 0, out_buffer, 132) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	output->Rank = pop_uint32_t(&p);
	for(i=0; i<128; i++) output->Reserved0[i] = pop_uint8_t(&p);
	return result_ok;
}

result_t ivm_get_measurement_settings(device_t handle, ivm_measurement_settings_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t out_buffer[48];
	memset(out_buffer, 0, 48);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "gmes", NULL, 0, out_buffer, 48) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	output->SamplingRate = pop_float(&p);
	output->ProbeSignalFrequency = pop_float(&p);
	output->NumberPoints = pop_uint32_t(&p);
	output->NumberChargePoints = pop_uint32_t(&p);
	output->MaxVoltage = pop_float(&p);
	for(i=0; i<4; i++) output->Reserved0[i] = pop_uint8_t(&p);
	output->CurrentSensorMode = pop_uint32_t(&p);
	output->OutputMode = pop_uint8_t(&p);
	for(i=0; i<19; i++) output->Reserved1[i] = pop_uint8_t(&p);
	return result_ok;
}

result_t ivm_set_measurement_settings(device_t handle, ivm_measurement_settings_t *input)
{
	uint8_t *p;
	unsigned int i;
	uint8_t in_buffer[48];
	memset(in_buffer, 0, 48);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	p = in_buffer;
	push_float(&p, input->SamplingRate);
	push_float(&p, input->ProbeSignalFrequency);
	push_uint32_t(&p, input->NumberPoints);
	push_uint32_t(&p, input->NumberChargePoints);
	push_float(&p, input->MaxVoltage);
	for(i=0; i<4; i++) push_uint8_t(&p, input->Reserved0[i]);
	push_uint32_t(&p, input->CurrentSensorMode);
	push_uint8_t(&p, input->OutputMode);
	for(i=0; i<19; i++) push_uint8_t(&p, input->Reserved1[i]);
	if(urpc_device_send_request(device, "smes", in_buffer, 48, NULL, 0) != urpc_result_ok)
	{
	    return result_error;
	}
	return result_ok;
}

result_t ivm_get_calibration_settings(device_t handle, ivm_calibration_settings_t *output)
{
	uint8_t *p;
	unsigned int i;
	uint8_t out_buffer[112];
	memset(out_buffer, 0, 112);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	if(urpc_device_send_request(device, "gcal", NULL, 0, out_buffer, 112) != urpc_result_ok)
	{
	    return result_error;
	}
	p = out_buffer;
	for(i=0; i<8; i++) output->Reserved0[i] = pop_uint8_t(&p);
	output->ADCMult = pop_float(&p);
	output->ADCVOffset = pop_float(&p);
	output->CurrentSense1Mult = pop_float(&p);
	output->CurrentSense1Offset = pop_float(&p);
	output->CurrentSense2Mult = pop_float(&p);
	output->CurrentSense2Offset = pop_float(&p);
	output->CurrentSense3Mult = pop_float(&p);
	output->CurrentSense3Offset = pop_float(&p);
	for(i=0; i<72; i++) output->Reserved1[i] = pop_uint8_t(&p);
	return result_ok;
}

result_t ivm_set_calibration_settings(device_t handle, ivm_calibration_settings_t *input)
{
	uint8_t *p;
	unsigned int i;
	uint8_t in_buffer[112];
	memset(in_buffer, 0, 112);
	urpc_device_handle_t device;
	if(handle < 0)
	{
	    return result_error;
	}
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	}
	p = in_buffer;
	for(i=0; i<8; i++) push_uint8_t(&p, input->Reserved0[i]);
	push_float(&p, input->ADCMult);
	push_float(&p, input->ADCVOffset);
	push_float(&p, input->CurrentSense1Mult);
	push_float(&p, input->CurrentSense1Offset);
	push_float(&p, input->CurrentSense2Mult);
	push_float(&p, input->CurrentSense2Offset);
	push_float(&p, input->CurrentSense3Mult);
	push_float(&p, input->CurrentSense3Offset);
	for(i=0; i<72; i++) push_uint8_t(&p, input->Reserved1[i]);
	if(urpc_device_send_request(device, "scal", in_buffer, 112, NULL, 0) != urpc_result_ok)
	{
	    return result_error;
	}
	return result_ok;
}

result_t ivm_close_device(device_t *handle_ptr)
{
	if(handle_ptr == NULL)
	{
	    return result_error;
	}
	device_t handle = *handle_ptr;
	if(handle < 0)
	{
	    return result_error;
	}
	urpc_device_handle_t device;
	{
	    std::lock_guard<std::mutex> lock(impl_by_handle_mutex);
	    try
	    {
	        device = impl_by_handle.at(handle);
	    }
	    catch(const std::out_of_range &)
	    {
	        return result_error;
	    }
	    impl_by_handle.erase(handle);
	}
	if(urpc_device_destroy(device) != urpc_result_ok)
	{
	    return result_error;
	}
	return result_ok;
}

